@@ -1,3 +1,4 @@
+import argparse
 import io
 import json
 import os
@@ -112,6 +113,41 @@ def connect_to_robot(
     }
 
 
+@mcp.tool(description="Detect the ROS version and distribution via rosbridge.")
+def detect_ros_version() -> dict:
+    """
+    Detects the ROS version and distro via rosbridge WebSocket.
+    Returns:
+        dict: {'version': <version or '1'>, 'distro': <distro>} or error info.
+    """
+    # Try ROS2 detection
+    ros2_request = {
+        "op": "call_service",
+        "id": "ros2_version_check",
+        "service": "/rosapi/get_ros_version",
+        "args": {},
+    }
+    with ws_manager:
+        response = ws_manager.request(ros2_request)
+        values = response.get("values") if response else None
+        if isinstance(values, dict) and "version" in values:
+            return {"version": values.get("version"), "distro": values.get("distro")}
+        # Fallback to ROS1 detection
+        ros1_request = {
+            "op": "call_service",
+            "id": "ros1_distro_check",
+            "service": "/rosapi/get_param",
+            "args": {"name": "/rosdistro"},
+        }
+        response = ws_manager.request(ros1_request)
+        value = response.get("values") if response else None
+        if value:
+            distro = value.get("value") if isinstance(value, dict) else value
+            distro_clean = str(distro).strip('"').replace("\\n", "").replace("\n", "")
+            return {"version": "1", "distro": distro_clean}
+        return {"error": "Could not detect ROS version"}
+
+
 @mcp.tool(description=("Fetch available topics from the ROS bridge.\nExample:\nget_topics()"))
 def get_topics() -> dict:
     """
@@ -357,6 +393,96 @@ def get_subscribers_for_topic(topic: str) -> dict:
         return {"error": f"Failed to get subscribers for topic {topic}"}
 
 
+@mcp.tool(
+    description=(
+        "Get comprehensive information about all ROS topics including publishers, subscribers, and message types. Note that this may take time to execute when three are a large number of topics since it queries each one by one under the hood. \n"
+        "Example:\n"
+        "inspect_all_topics()"
+    )
+)
+def inspect_all_topics() -> dict:
+    """
+    Get comprehensive information about all ROS topics including publishers, subscribers, and message types.
+
+    Returns:
+        dict: Contains detailed information about all topics including:
+            - Topic names and message types
+            - Publishers for each topic
+            - Subscribers for each topic
+            - Connection counts and statistics
+    """
+    # First get all topics
+    topics_message = {
+        "op": "call_service",
+        "service": "/rosapi/topics",
+        "type": "rosapi/Topics",
+        "args": {},
+        "id": "inspect_all_topics_request_1",
+    }
+
+    with ws_manager:
+        topics_response = ws_manager.request(topics_message)
+
+        if not topics_response or "values" not in topics_response:
+            return {"error": "Failed to get topics list"}
+
+        topics = topics_response["values"].get("topics", [])
+        types = topics_response["values"].get("types", [])
+        topic_details = {}
+
+        # Get details for each topic
+        topic_errors = []
+        for i, topic in enumerate(topics):
+            # Get topic type
+            topic_type = types[i] if i < len(types) else "unknown"
+
+            # Get publishers for this topic
+            publishers_message = {
+                "op": "call_service",
+                "service": "/rosapi/publishers",
+                "type": "rosapi/Publishers",
+                "args": {"topic": topic},
+                "id": f"get_publishers_{topic.replace('/', '_')}",
+            }
+
+            publishers_response = ws_manager.request(publishers_message)
+            publishers = []
+            if publishers_response and "values" in publishers_response:
+                publishers = publishers_response["values"].get("publishers", [])
+            elif publishers_response and "error" in publishers_response:
+                topic_errors.append(f"Topic {topic} publishers: {publishers_response['error']}")
+
+            # Get subscribers for this topic
+            subscribers_message = {
+                "op": "call_service",
+                "service": "/rosapi/subscribers",
+                "type": "rosapi/Subscribers",
+                "args": {"topic": topic},
+                "id": f"get_subscribers_{topic.replace('/', '_')}",
+            }
+
+            subscribers_response = ws_manager.request(subscribers_message)
+            subscribers = []
+            if subscribers_response and "values" in subscribers_response:
+                subscribers = subscribers_response["values"].get("subscribers", [])
+            elif subscribers_response and "error" in subscribers_response:
+                topic_errors.append(f"Topic {topic} subscribers: {subscribers_response['error']}")
+
+            topic_details[topic] = {
+                "type": topic_type,
+                "publishers": publishers,
+                "subscribers": subscribers,
+                "publisher_count": len(publishers),
+                "subscriber_count": len(subscribers),
+            }
+
+        return {
+            "total_topics": len(topics),
+            "topics": topic_details,
+            "topic_errors": topic_errors,  # Include any errors encountered during inspection
+        }
+
+
 @mcp.tool(
     description=(
         "Subscribe to a ROS topic and return the first message received.\n"
@@ -451,7 +577,7 @@ def subscribe_once(
                 ws_manager.send(unsubscribe_msg)
                 if "Image" in msg_type:
                     return {
-                        "message": "Image received successfully and saved in the MCP server. Run the 'analyze_image' tool to analyze it"
+                        "message": "Image received successfully and saved in the MCP server. Run the 'analyze_previously_received_image' tool to analyze it"
                     }
                 else:
                     return {"msg": msg_data.get("msg", {})}
@@ -933,11 +1059,11 @@ def get_service_providers(service: str) -> dict:
     if not service or not service.strip():
         return {"error": "Service name cannot be empty"}
 
-    # rosbridge service call to get service providers
+    # rosbridge service call to get service providers (using service_node like inspect_all_services)
     message = {
         "op": "call_service",
-        "service": "/rosapi/service_providers",
-        "type": "rosapi/ServiceProviders",
+        "service": "/rosapi/service_node",
+        "type": "rosapi/ServiceNode",
         "args": {"service": service},
         "id": f"get_service_providers_request_{service.replace('/', '_')}",
     }
@@ -946,17 +1072,34 @@ def get_service_providers(service: str) -> dict:
     with ws_manager:
         response = ws_manager.request(message)
 
-    # Return service providers if present
-    if response and "values" in response:
-        providers = response["values"].get("providers", [])
-        return {"service": service, "providers": providers, "provider_count": len(providers)}
+    # Return service providers if present (using same logic as inspect_all_services)
+    providers = []
+
+    # Handle different response formats safely
+    if response and isinstance(response, dict):
+        if "values" in response:
+            node = response["values"].get("node", "")
+            if node:
+                providers = [node]
+        elif "result" in response:
+            node = response["result"].get("node", "")
+            if node:
+                providers = [node]
+        elif "error" in response:
+            return {"error": f"Service call failed: {response['error']}"}
+    elif response is False:
+        return {"error": f"No response received for service {service}"}
+    elif response is True:
+        return {"error": f"Unexpected boolean response for service {service}"}
     else:
         return {"error": f"Failed to get providers for service {service}"}
 
+    return {"service": service, "providers": providers, "provider_count": len(providers)}
+
 
 @mcp.tool(
     description=(
-        "Get comprehensive information about all services including types and providers.\n"
+        "Get comprehensive information about all services including types and providers. Note that this may take time to execute when three are a large number of services since it queries each one by one under the hood. \n"
         "Example:\n"
         "inspect_all_services()"
     )
@@ -1006,21 +1149,36 @@ def inspect_all_services() -> dict:
             elif type_response and "error" in type_response:
                 service_errors.append(f"Service {service}: {type_response['error']}")
 
-            # Get service providers
-            providers_message = {
+            # Get service provider (using service_node instead of service_providers)
+            provider_message = {
                 "op": "call_service",
-                "service": "/rosapi/service_providers",
-                "type": "rosapi/ServiceProviders",
+                "service": "/rosapi/service_node",
+                "type": "rosapi/ServiceNode",
                 "args": {"service": service},
-                "id": f"get_providers_{service.replace('/', '_')}",
+                "id": f"get_provider_{service.replace('/', '_')}",
             }
 
-            providers_response = ws_manager.request(providers_message)
+            provider_response = ws_manager.request(provider_message)
             providers = []
-            if providers_response and "values" in providers_response:
-                providers = providers_response["values"].get("providers", [])
-            elif providers_response and "error" in providers_response:
-                service_errors.append(f"Service {service} providers: {providers_response['error']}")
+
+            # Handle different response formats safely
+            if provider_response and isinstance(provider_response, dict):
+                if "values" in provider_response:
+                    node = provider_response["values"].get("node", "")
+                    if node:
+                        providers = [node]
+                elif "result" in provider_response:
+                    node = provider_response["result"].get("node", "")
+                    if node:
+                        providers = [node]
+                elif "error" in provider_response:
+                    service_errors.append(
+                        f"Service {service} provider: {provider_response['error']}"
+                    )
+            elif provider_response is False:
+                service_errors.append(f"Service {service} provider: No response received")
+            elif provider_response is True:
+                service_errors.append(f"Service {service} provider: Unexpected boolean response")
 
             service_details[service] = {
                 "type": service_type,
@@ -1118,6 +1276,202 @@ def call_service(
         }
 
 
+@mcp.tool(description=("Get list of all currently running ROS nodes.\nExample:\nget_nodes()"))
+def get_nodes() -> dict:
+    """
+    Get list of all currently running ROS nodes.
+
+    Returns:
+        dict: Contains list of all active nodes,
+            or a message string if no nodes are found.
+    """
+    # rosbridge service call to get node list
+    message = {
+        "op": "call_service",
+        "service": "/rosapi/nodes",
+        "type": "rosapi/Nodes",
+        "args": {},
+        "id": "get_nodes_request_1",
+    }
+
+    # Request node list from rosbridge
+    with ws_manager:
+        response = ws_manager.request(message)
+
+    # Check for service response errors first
+    if response and "result" in response and not response["result"]:
+        # Service call failed - return error with details from values
+        error_msg = response.get("values", {}).get("message", "Service call failed")
+        return {"error": f"Service call failed: {error_msg}"}
+
+    # Return node info if present
+    if response and "values" in response:
+        nodes = response["values"].get("nodes", [])
+        return {"nodes": nodes, "node_count": len(nodes)}
+    else:
+        return {"warning": "No nodes found"}
+
+
+@mcp.tool(
+    description=(
+        "Get detailed information about a specific node including its publishers, subscribers, and services.\n"
+        "Example:\n"
+        "get_node_details('/turtlesim')"
+    )
+)
+def get_node_details(node: str) -> dict:
+    """
+    Get detailed information about a specific node including its publishers, subscribers, and services.
+
+    Args:
+        node (str): The node name (e.g., '/turtlesim')
+
+    Returns:
+        dict: Contains detailed node information including publishers, subscribers, and services,
+            or an error message if node doesn't exist.
+    """
+    # Validate input
+    if not node or not node.strip():
+        return {"error": "Node name cannot be empty"}
+
+    result = {
+        "node": node,
+        "publishers": [],
+        "subscribers": [],
+        "services": [],
+        "publisher_count": 0,
+        "subscriber_count": 0,
+        "service_count": 0,
+    }
+
+    # rosbridge service call to get node details
+    message = {
+        "op": "call_service",
+        "service": "/rosapi/node_details",
+        "type": "rosapi/NodeDetails",
+        "args": {"node": node},
+        "id": f"get_node_details_{node.replace('/', '_')}",
+    }
+
+    # Request node details from rosbridge
+    with ws_manager:
+        response = ws_manager.request(message)
+
+    # Check for service response errors first
+    if response and "result" in response and not response["result"]:
+        # Service call failed - return error with details from values
+        error_msg = response.get("values", {}).get("message", "Service call failed")
+        return {"error": f"Service call failed: {error_msg}"}
+
+    # Extract data from the response
+    if response and "values" in response:
+        values = response["values"]
+        # Extract publishers, subscribers, and services from the response
+        # Note: rosapi uses "publishing" and "subscribing" field names
+        publishers = values.get("publishing", [])
+        subscribers = values.get("subscribing", [])
+        services = values.get("services", [])
+
+        result["publishers"] = publishers
+        result["subscribers"] = subscribers
+        result["services"] = services
+        result["publisher_count"] = len(publishers)
+        result["subscriber_count"] = len(subscribers)
+        result["service_count"] = len(services)
+
+    # Check if we got any data
+    if not result["publishers"] and not result["subscribers"] and not result["services"]:
+        return {"error": f"Node {node} not found or has no details available"}
+
+    return result
+
+
+@mcp.tool(
+    description=(
+        "Get comprehensive information about all ROS nodes including their publishers, subscribers, and services.\n"
+        "Example:\n"
+        "inspect_all_nodes()"
+    )
+)
+def inspect_all_nodes() -> dict:
+    """
+    Get comprehensive information about all ROS nodes including their publishers, subscribers, and services.
+
+    Returns:
+        dict: Contains detailed information about all nodes including:
+            - Node names and details
+            - Publishers for each node
+            - Subscribers for each node
+            - Services provided by each node
+            - Connection counts and statistics
+    """
+    # First get all nodes
+    nodes_message = {
+        "op": "call_service",
+        "service": "/rosapi/nodes",
+        "type": "rosapi/Nodes",
+        "args": {},
+        "id": "inspect_all_nodes_request_1",
+    }
+
+    with ws_manager:
+        nodes_response = ws_manager.request(nodes_message)
+
+        if not nodes_response or "values" not in nodes_response:
+            return {"error": "Failed to get nodes list"}
+
+        nodes = nodes_response["values"].get("nodes", [])
+        node_details = {}
+
+        # Get details for each node
+        node_errors = []
+        for node in nodes:
+            # Get node details (publishers, subscribers, services)
+            node_details_message = {
+                "op": "call_service",
+                "service": "/rosapi/node_details",
+                "type": "rosapi/NodeDetails",
+                "args": {"node": node},
+                "id": f"get_node_details_{node.replace('/', '_')}",
+            }
+
+            node_details_response = ws_manager.request(node_details_message)
+
+            if node_details_response and "values" in node_details_response:
+                values = node_details_response["values"]
+                # Extract publishers, subscribers, and services from the response
+                # Note: rosapi uses "publishing" and "subscribing" field names
+                publishers = values.get("publishing", [])
+                subscribers = values.get("subscribing", [])
+                services = values.get("services", [])
+
+                node_details[node] = {
+                    "publishers": publishers,
+                    "subscribers": subscribers,
+                    "services": services,
+                    "publisher_count": len(publishers),
+                    "subscriber_count": len(subscribers),
+                    "service_count": len(services),
+                }
+            elif (
+                node_details_response
+                and "result" in node_details_response
+                and not node_details_response["result"]
+            ):
+                error_msg = node_details_response.get("values", {}).get(
+                    "message", "Service call failed"
+                )
+                node_errors.append(f"Node {node}: {error_msg}")
+            else:
+                node_errors.append(f"Node {node}: Failed to get node details")
+
+        return {
+            "total_nodes": len(nodes),
+            "nodes": node_details,
+            "node_errors": node_errors,  # Include any errors encountered during inspection
+        }
+
+
 ## ############################################################################################## ##
 ##
 ##                       NETWORK DIAGNOSTICS
@@ -1154,27 +1508,6 @@ def ping_robot(ip: str, port: int, ping_timeout: float = 2.0, port_timeout: floa
 ##                      IMAGE ANALYSIS
 ##
 ## ############################################################################################## ##
-@mcp.tool(
-    description=(
-        "First, subscribe to an Image topic using 'subscribe_once' to save an image.\n"
-        "Then, use this tool to analyze the saved image\n"
-    )
-)
-def analyze_previously_received_image():
-    """
-    Analyze the received image.
-
-    This tool loads the previously saved image from './camera/received_image.jpeg'
-    (which must have been created by 'parse_image' or 'subscribe_once'), and converts
-    it into an MCP-compatible ImageContent format so that the LLM can interpret it.
-    """
-    path = "./camera/received_image.jpeg"
-    if not os.path.exists(path):
-        return {"error": "No previously received image found at ./camera/received_image.jpeg"}
-    image = PILImage.open(path)
-    return _encode_image_to_imagecontent(image)
-
-
 def _encode_image_to_imagecontent(image):
     """
     Encodes a PIL Image to a format compatible with ImageContent.
@@ -1192,7 +1525,74 @@ def _encode_image_to_imagecontent(image):
     return img_obj.to_image_content()
 
 
-if __name__ == "__main__":
+@mcp.tool(
+    description=(
+        "First, subscribe to an Image topic using 'subscribe_once' to save an image.\n"
+        "Then, use this tool to analyze the saved image\n"
+    )
+)
+def analyze_previously_received_image():
+    """
+    Analyze the previously received image saved at ./camera/received_image.jpeg
+
+    This tool loads the previously saved image from './camera/received_image.jpeg'
+    (which must have been created by 'parse_image' or 'subscribe_once'), and converts
+    it into an MCP-compatible ImageContent format so that the LLM can interpret it.
+    """
+    path = "./camera/received_image.jpeg"
+    if not os.path.exists(path):
+        return {"error": "No image found at ./camera/received_image.jpeg"}
+    img = PILImage.open(path)
+    return _encode_image_to_imagecontent(img)
+
+
+def parse_arguments():
+    """Parse command line arguments for MCP server configuration."""
+    parser = argparse.ArgumentParser(
+        description="ROS MCP Server - Connect to ROS robots via MCP protocol",
+        formatter_class=argparse.RawDescriptionHelpFormatter,
+        epilog="""
+Examples:
+  python server.py                                    # Use stdio transport (default)
+  python server.py --transport http --host 0.0.0.0 --port 9000
+  python server.py --transport streamable-http --host 127.0.0.1 --port 8080
+        """,
+    )
+
+    parser.add_argument(
+        "--transport",
+        choices=["stdio", "http", "streamable-http", "sse"],
+        default="stdio",
+        help="MCP transport protocol to use (default: stdio)",
+    )
+
+    parser.add_argument(
+        "--host",
+        default="127.0.0.1",
+        help="Host address for HTTP-based transports (default: 127.0.0.1)",
+    )
+
+    parser.add_argument(
+        "--port",
+        type=int,
+        default=9000,
+        help="Port number for HTTP-based transports (default: 9000)",
+    )
+
+    return parser.parse_args()
+
+
+def main():
+    """Main entry point for the MCP server console script."""
+    # Parse command line arguments
+    args = parse_arguments()
+
+    # Update global variables with parsed arguments
+    global MCP_TRANSPORT, MCP_HOST, MCP_PORT
+    MCP_TRANSPORT = args.transport.lower()
+    MCP_HOST = args.host
+    MCP_PORT = args.port
+
     if MCP_TRANSPORT == "stdio":
         # stdio doesn't need host/port
         mcp.run(transport="stdio")
@@ -1212,3 +1612,7 @@ if __name__ == "__main__":
             f"Unsupported MCP_TRANSPORT={MCP_TRANSPORT!r}. "
             "Use 'stdio', 'http', or 'streamable-http'."
         )
+
+
+if __name__ == "__main__":
+    main()
